#include "pitch_track.h"
#include <iostream>
#include "Windows.h"

int plltrack_set(pitch_track_CSOUND *csound, PLLTRACK *p)
{
    int i;
    p->x1 = p->cos_x = p->sin_x = 0.0;
    p->x2 = 1.0;
    p->klpf_o = p->klpfQ_o = p->klf_o = p->khf_o = 0.0;
    update_coefs(csound,10.0, 0.0, &p->fils[4], LP1);
    p->ace = p->xce = 0.0;
    for (i=0; i < 6; i++)
        p->fils[i].del1 = p->fils[i].del2 = 0.0;

    return 0;
}

int plltrack_perf(pitch_track_CSOUND *csound, PLLTRACK *p)
{
    int ksmps, i, k;
    MYFLT _0dbfs;
    double a0[6], a1[6], a2[6], b1[6], b2[6];
    double *mem1[6], *mem2[6];
    double *ace, *xce;
    double *cos_x, *sin_x, *x1, *x2;
    double scal,esr;
    BIQUAD *biquad = p->fils;
    MYFLT *asig=p->asig,klpf,klpfQ,klf,khf;
    MYFLT kd=p->kd;
    MYFLT kthresh;
    MYFLT *freq=p->freq.data();
    MYFLT *lock=p->lock.data();
    klpf = p->klpf;
    klpfQ = p->klpfQ;
    klf = p->klf;
    khf = p->khf;
    kthresh = p->kthresh;

    _0dbfs = csound->e0dbfs;
    ksmps = csound->GetKsmps();
    esr = csound->GetSr();
    scal = 2.0*csound->pidsr;

    if (p->khf_o != khf) {
        update_coefs(csound, khf, 0.0, &biquad[0], LP1);
        update_coefs(csound, khf, 0.0, &biquad[1], LP1);
        update_coefs(csound, khf, 0.0, &biquad[2], LP1);
        p->khf_o = khf;
    }

    if (p->klf_o != klf) {
        update_coefs(csound, klf, 0.0, &biquad[3], HP);
        p->klf_o = klf;
    }

    if (p->klpf_o != klpf || p->klpfQ_o != klpfQ ) {
        update_coefs(csound, klpf, klpfQ, &biquad[5], LP2);
        p->klpf_o = klpf; p->klpfQ_o = klpfQ;
    }



    for(k=0; k < 6; k++) {
        a0[k] = biquad[k].a0;
        a1[k] = biquad[k].a1;
        a2[k] = biquad[k].a2;
        b1[k] = biquad[k].b1;
        b2[k] = biquad[k].b2;
        mem1[k] = &(biquad[k].del1);
        mem2[k] = &(biquad[k].del2);
    }

    cos_x = &p->cos_x;
    sin_x = &p->sin_x;
    x1 = &p->x1;
    x2 = &p->x2;
    xce = &p->xce;
    ace = &p->ace;

    for (i=0; i < ksmps; i++){
        double input = (double) (asig[i]/_0dbfs), env;
        double w, y, icef = 0.99, fosc, xd, c, s, oc;

        /* input stage filters */
        for (k=0; k < 4 ; k++){
            w =  input - *(mem1[k])*b1[k] - *(mem2[k])*b2[k];
            y  = w*a0[k] + *(mem1[k])*a1[k] + *(mem2[k])*a2[k];
            *(mem2[k]) = *(mem1[k]);
            *(mem1[k]) = w;
            input = y;
        }

        /* envelope extraction */
        w =  fabs(input) - *(mem1[k])*b1[k] - *(mem2[k])*b2[k];
        y  = w*a0[k] + *(mem1[k])*a1[k] + *(mem2[k])*a2[k];
        *(mem2[k]) = *(mem1[k]);
        *(mem1[k]) = w;
        env = y;
        k++;

        /* constant envelope */
        if (env > kthresh)
            input /= env;
        else input = 0.0;

        /*post-ce filter */
        *ace = (1.-icef)*(input + *xce)/2. + *ace*icef;
        *xce = input;

        /* PLL */
        xd =  *cos_x * (*ace) * kd * esr;
        w =  xd - *(mem1[k])*b1[k] - *(mem2[k])*b2[k];
        y  = w*a0[k] + *(mem1[k])*a1[k] + *(mem2[k])*a2[k];
        *(mem2[k]) = *(mem1[k]);
        *(mem1[k]) = w;
        freq[i] = fabs(2*y);
        lock[i] = *ace * (*sin_x);
        fosc = y + xd;

        /* quadrature osc */
        *sin_x = *x1;
        *cos_x = *x2;
        oc = fosc*scal;
        c = cos(oc);  s = sin(oc);
        *x1 = *sin_x*c + *cos_x*s;
        *x2 = -*sin_x*s + *cos_x*c;

    }
    return 0;
}

void update_coefs(pitch_track_CSOUND *csound, double fr, double Q, BIQUAD *biquad, int TYPE)
{
    double k, ksq, div, ksqQ;

    switch(TYPE){
    case LP2:
        k = tan(fr*csound->pidsr);
        ksq = k*k;
        ksqQ = ksq*Q;
        div = ksqQ+k+Q;
        biquad->b1 = (2*Q*(ksq-1.))/div;
        biquad->b2 = (ksqQ-k+Q)/div;
        biquad->a0 = ksqQ/div;
        biquad->a1 = 2*biquad->a0;
        biquad->a2 = biquad->a0;
        break;

    case LP1:
        k = 1.0/tan(csound->pidsr*fr);
        ksq = k*k;
        biquad->a0 = 1.0 / ( 1.0 + ROOT2 * k + ksq);
        biquad->a1 = 2.0*biquad->a0;
        biquad->a2 = biquad->a0;
        biquad->b1 = 2.0 * (1.0 - ksq) * biquad->a0;
        biquad->b2 = ( 1.0 - ROOT2 * k + ksq) * biquad->a0;
        break;

    case HP:
        k = tan(csound->pidsr*fr);
        ksq = k*k;
        biquad->a0 = 1.0 / ( 1.0 + ROOT2 * k + ksq);
        biquad->a1 = -2.*biquad->a0;
        biquad->a2 = biquad->a0;
        biquad->b1 = 2.0 * (ksq - 1.0) * biquad->a0;
        biquad->b2 = ( 1.0 - ROOT2 * k + ksq) * biquad->a0;
        break;
    }

}

pll_pitch_tracker::pll_pitch_tracker()
{
    m_params.push_back(hourglass_plugin_parameter("Feedback","%",0.01,1.0,0.5,[this](double x){ pll.kd=x; }));
    m_params.push_back(hourglass_plugin_parameter("Loop frequency","Hz",1.0,30.0,20.0,[this](double x){ pll.klpf=x; }));
    m_params.push_back(hourglass_plugin_parameter("Loop Q","units",0.01,1.0,0.333,[this](double x){ pll.klpfQ=x; }));
    auto par_setter=[this](double x)
    {
        pll.kthresh=exp(x*0.11512925464970228420089957273422);
    };
    m_params.push_back(hourglass_plugin_parameter("Threshold","dB",-96.0,-3.0,-60.0,par_setter));
}

bool pll_pitch_tracker::process(double* audio_in, size_t frames, double sr,std::vector<pitch_info>& output)
{
    if (output.size()<frames)
        output.resize(frames);
    csound.ksmps=frames;
    csound.sr=sr;
    csound.pidsr=3.141592654/sr;
    pll.asig=audio_in;
    plltrack_perf(&csound,&pll);
    for (int i=0;i<frames;i++)
    {
        output[i].hz=pll.freq[i];
        output[i].qual=pll.lock[i];
    }
    return true;
}


IHourGlass_pitch_tracker* create_pitch_tracker()
{
    //OutputDebugStringA("CREATE PITCH TRACKER\n");
    try
    {
        return new pll_pitch_tracker;
    }
    catch (std::exception& e)
    {
        std::cout << e.what() << "\n";
        return nullptr;
    }

}

void destroy_pitch_tracker(IHourGlass_pitch_tracker* p)
{
    //char buf[200];
    //sprintf(buf,"DESTROY PITCH TRACKER %d\n",p);
    //OutputDebugStringA(buf);
    delete p;
}
